"""
:mod:`zsl.interface.webservice.utils.response_headers`
------------------------------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals
from flask import Response
from flask.helpers import make_response
from functools import wraps

from zsl import inject, Zsl, Config
from zsl.task.task_decorator import CrossdomainWebTaskResponder


@inject(config=Config)
def append_crossdomain(response, config):
    # type: (Response, Config)->None
    """
    Adds the default crossdomain headers.
    Uses the :class:`zsl.task.task_decorator.CrossdomainWebTaskResponder`
    to generate the headers.

    :param response: Current web response.
    :param config:  Current configuration.
    """
    # The CORS has already been setup.
    if 'Access-Control-Allow-Origin' in response.headers:
        return
    CrossdomainWebTaskResponder().respond(response)


@inject(app=Zsl)
def append_asl(response, app):
    response.headers['ZSL-Flask-Layer'] = app.get_version()


def append_cache(response):
    response.headers['Cache-Control'] = 'no-cache'


def append_all(response):
    append_asl(response)
    append_cache(response)
    append_crossdomain(response)


def append_headers(f):
    """
    Appends all the web headers:
      * ZSL version and information,
      * default CORS if not already set up,
      * cache.

    :param f: The decorated function.
    :return: The function which appends the web headers.
    """
    @wraps(f)
    def _response_decorator(*args, **kwargs):
        r = f(*args, **kwargs)
        response = r if isinstance(r, Response) else make_response(r)
        append_all(response)
        return response

    return _response_decorator



"""
:mod:`zsl.utils.import_helper`
------------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
import importlib


def fetch_class(full_class_name):
    """Fetches the given class.

    :param string full_class_name: Name of the class to be fetched.
    """
    (module_name, class_name) = full_class_name.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)



"""
:mod:`zsl.utils.request_helper`

Helper module for working with request.

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals
from future.utils import viewitems


def args_to_dict(args):
    """Converts request arguments to a simple dictionary. Uses only the first
    value.

    :param args: flask's request args object
    :type args: request.args
    :return: a simple dict from args
    :rtype: dict
    """
    return args.to_dict(flat=True)



"""
:mod:`zsl.testing.db`
---------------------
This module allows for database unit testing. For how to use the database
testing in practice, a sample, refer to :ref:`unit-testing-db`.

The module works in the following way (methods setUp, tearDown):
1. Each test runs in a single transaction.
2. This transaction is always called a rollback.

All the tests are run in a single parent transaction (setUpClass,
tearDownClass):
1. In general initialization phase the session/transaction is created
and it is kept during all the testing. Also the database schema is created.
2. After this the transaction is called rollback.

This means that the tests may be conducted in the in the memory database
or a persistent one which is kept clean.

The module provides class :class:`.TestSessionFactory` - it always returns
the same session. Also one should add :class:`.DbTestModule` to the test
container when creating Zsl instance, see :ref:`unit-testing-zsl-instance`.
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import *

import logging
from injector import Module, provides, singleton
from sqlalchemy.engine import Engine
from sqlalchemy.orm.session import Session

from zsl import inject
from zsl.application.modules.alchemy_module import TransactionHolderFactory, \
    TransactionHolder
from zsl.db.model.sql_alchemy import metadata
from zsl.service.service import SessionFactory


class TestSessionFactory(SessionFactory):
    """Factory always returning the single test transaction."""
    _test_session = None

    @inject(engine=Engine)
    def create_test_session(self, engine):
        # type: (Engine) -> Session
        assert TestSessionFactory._test_session is None
        metadata.bind = engine
        metadata.create_all(engine)
        logging.getLogger(__name__).debug("Create test session")
        TestSessionFactory._test_session = self._session_holder()
        TestSessionFactory._test_session.autoflush = True
        TestSessionFactory._test_session.begin(subtransactions=True)
        assert TestSessionFactory._test_session is not None
        return TestSessionFactory._test_session

    def create_session(self):
        logging.getLogger(__name__).debug("Create session")
        assert TestSessionFactory._test_session is not None
        TestSessionFactory._test_session.begin_nested()
        return TestSessionFactory._test_session

    def close_test_session(self):
        TestSessionFactory._test_session.rollback()
        TestSessionFactory._test_session.close()
        TestSessionFactory._test_session = None


class TestTransactionHolderFactory(TransactionHolderFactory):
    def create_transaction_holder(self):
        return TransactionHolder()


class DbTestModule(Module):
    """Module fixing the :class:`zsl.service.service.SessionFactory`
    to our :class:`.TestSessionFactory`."""

    @provides(SessionFactory, scope=singleton)
    def get_session_factory(self):
        # type: ()->SessionFactory
        return TestSessionFactory()

    @provides(TestSessionFactory, scope=singleton)
    @inject(session_factory=SessionFactory)
    def get_test_session_factory(self, session_factory):
        # type: (SessionFactory)->SessionFactory
        return session_factory

    @provides(TransactionHolderFactory, scope=singleton)
    def provide_transaction_holder_factory(self):
        return TestTransactionHolderFactory()


class DbTestCase(object):
    """:class:`.DbTestCase` is a mixin to be used when testing with
    a database."""

    _session = None

    @inject(session_factory=TestSessionFactory)
    def setUp(self, session_factory):
        super(DbTestCase, self).setUp()
        # type: (TestSessionFactory)->None
        logging.getLogger(__name__).debug("DbTestCase.setUp")
        session_factory.create_test_session()

    @inject(session_factory=TestSessionFactory)
    def tearDown(self, session_factory):
        # type: (TestSessionFactory)->None
        # This will return the same transaction/session
        # as the one used in setUp.
        logging.getLogger(__name__).debug("DbTestCase.tearDown")
        session_factory.close_test_session()
        super(DbTestCase, self).tearDown()


IN_MEMORY_DB_SETTINGS = {
    'DATABASE_URI': 'sqlite:///:memory:',
    'DATABASE_ENGINE_PROPS': {},
    'JSON_AS_ASCII': False
}
